#include <hesiod.h>
#include <pwd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <syslog.h>
#include <errno.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#ifdef linux
#include <sys/vfs.h>
#endif
#ifdef __osf__
#include <sys/mount.h>
#endif

/*
 * worker - called by sserver 
 *        - inputs name of user, ip of his workstation
 *	  - gets hesiod home directory, then any shared directores
 *        - makes mods to exports file/s and hups daemon/s
 *	  - returns list of fstab entries that client machine 
 *          can now mount.
 */

void worker(char *name, char *ip, char *repbuf){

    struct  passwd *pwd;
    char    dir[1024];
    FILE    *f;
    char    buf[1024], *pd;
    int     now;
    void    *context;
    char    huplist[1024];

    umask(022);
    huplist[0] = '\0';

    /* start with a little gratituitous info */
    now = time((time_t *) 0);
    sprintf(repbuf, 
        "You are %s on %s at %syour fstab entries are\n", 
        name, 
        ip, 
        ctime(&now)
           );

    /* get user's home dir via hesiod */
    if ( hesiod_init(&context) ){
        syslog(LOG_ERR, "hes_init: %m");
	strcpy(repbuf,"sorry - hesiod initialization error - see helpdesk");
	return;
    }
    pwd = hesiod_getpwnam(context, name);
    if ( !pwd ) {
         syslog(LOG_ERR, "hes_getpwnam: %m");
         strcpy(repbuf,"sorry - you were not found in hesiod");
         return;
    }
    strcpy(dir,pwd->pw_dir);
    hesiod_free_passwd(context, pwd);
    hesiod_end(context);

    /* for each dir we call setup */
    if ( setup(name,dir,ip,repbuf,huplist) ) return;

    /* grep for this user in /cc/st1/.ses.shares and repeat as above */
    f = fopen("/cc/st1/.ses.shares", "r" );
    if ( f == (FILE *) NULL ) {
      if ( errno != ENOENT ) {
        syslog(LOG_ERR,"nfsses grep %m");
      }
    }
    else {
      while ( ( ! feof(f) ) && (fgets( buf, 1024, f) ) ) {
        /* for non blank, non comment lines */
        if ( ( strlen(buf) > 1 ) && ( buf[0] != '#' )) {
          buf[strlen(buf)-1] = '\0';
          pd = index(buf, ' ');
          if ( pd == (char *)NULL ) continue;
          *pd++ = '\0';
          while ( *pd == ' ') pd++;
	  if ( ! strcmp(buf, name) ) {
            if ( setup(name, pd, ip, repbuf, huplist) ) return;
	  }
        }
      }
      fclose(f);
    }
/*    printf("huplist is %s\n",huplist); */
    hupdaemons(huplist);

    return;
}

/*
 * hupdaemons 
 *	- input is a list of machines whose mountd and possibly nfsd daemons
 *        need to be hupped.
 *      - the list doe not have repeats in it, and will usually only have
 *	  one member.
 */

hupdaemons( char *huplist){

  char *p, *q;

  p = huplist;
  while ( strlen(p) ){
    q = p;
    while ( (*q != '\0' ) && (*q != ' ') ) q++;
    if ( *q == ' ') {
      *q = '\0';
      hupone(p);
      *q = ' ';
      q++;
      p = q;
    }
    else {
      hupone(p);
      p = q;
    }
  }
  return;
}

/* 
 * hupone
 *	- input name of machine
 *	- hup its daemon or daemons
 */

hupone(char *name){

  char hn[128];
  char progname[128], *p;
  int i, f, pid;
  char buf [16];
  struct stat s;

  gethostname(hn,sizeof(hn));
/*  printf("hupping %s, we are %s\n",name,hn); */
  /* if its us, we just hup our own daemons.  easy! */
  if ( ! strcmp(name,hn) ) {
    /* its us.  now are we linux or tu ? */
    if ( stat("/var/ase/mnt",&s) || ( ! (s.st_mode & S_IFDIR) ) ) {
      /* local non-ase system - use kill */
/*      printf("local kill\n"); */
#ifdef linux
      strcpy(progname,"rpc.nfsd");
      p = &progname[0];			/* ugly.  fix this */
      if ( kill_all(1,1,&p) ) {
        syslog(LOG_ERR,"killall failed");
      }

      strcpy(progname,"rpc.mountd");
      p = &progname[0];			/* ugly.  fix this */
      if ( kill_all(1,1,&p) ) {
        syslog(LOG_ERR,"killall2 failed");
      }
#else
      f = open("/var/run/mountd.pid",O_RDONLY);
      if ( f < 0 ) { syslog(LOG_ERR,"open mountd.pid %m"); return; }
      i = read(f,buf,10);
      if ( i < 0 ) { syslog(LOG_ERR,"read mountd.pid %m"); return; }
      i = sscanf(buf,"%d",&pid);
      if ( i < 1 ) { syslog(LOG_ERR,"scan mountd.pid %m"); return; }
      kill( pid, 1 );
#endif
    }
    else {
      /* local ASE system  -  read /var/run/mountd.pid and send it a hup */
      f = open("/var/run/mountd.pid",O_RDONLY);
      if ( f < 0 ) { syslog(LOG_ERR,"open mountd.pid %m"); return; }
      i = read(f,buf,10);
      if ( i < 0 ) { syslog(LOG_ERR,"read mountd.pid %m"); return; }
      i = sscanf(buf,"%d",&pid);
      if ( i < 1 ) { syslog(LOG_ERR,"scan mountd.pid %m"); return; }
      kill( pid, 1 );
    }
  }
/*  else { */
    /* some sort of ASE system - send the hup to the machine */
    /* actually we don't need to do this on TU systems!!! */
/*    printf("sending a HUP to %s\n",name); */

/*  } */
  return;
}

/*
 *  setup    
 *	- for this dir, get fstab info - e.g. name of m/c to mount from
 *	- then update its entry in expire file
 *	- then check and possibly insert it in exports file, and HUP it.
 *	- then return fstab info to repbuf
 */

int setup(char *name, char *dir, char *ip, char *repbuf, char *huplist){

    char host[128], hostdir[128], servmntpt[128];
    char hn[128];
    struct stat s;

    /* first make sure the dir we were passed really is a directory */
    if ( stat(dir,&s) || ( ! (s.st_mode & S_IFDIR) ) ) {
	strcat(repbuf, "\n sorry - call helpdesk - directory ");
	strcat(repbuf, dir);
	strcat(repbuf, " not found by ");
        gethostname(hn,sizeof(hn));
        strcat(repbuf, hn);
	return(0);
    }

    if ( getmountinfo(dir,host,hostdir,servmntpt) ) {
	strcat(repbuf,"\n see helpdesk - error getting mount info\n");
        return(-1);
    }

    if ( fixexpire( name, ip, dir, servmntpt) ) {
	strcat(repbuf,"\n see helpdesk - error setting expiry info\n");
	return(-1);
    }

    if ( fixexports( ip, host, hostdir, servmntpt, huplist) ) {
	strcat(repbuf,"\n see helpdesk - error setting export info\n");
	return(-1);
    }

    strcat(repbuf,host);
    strcat(repbuf,":");
    strcat(repbuf,hostdir);
    strcat(repbuf," ");
    strcat(repbuf,dir);
    strcat(repbuf," nfs rw 0 0\n");
    return(0);
}

int fixexports( char *ip, char *host, char *hostdir, char *servmntpt, char *huplist) {

  char lockfile[128], expfile[128], newfile[128];
  FILE *f, *nf;
  char buf[1024], old[4096], new[4096];
  char *p, *q, csave, ipsav[20];
  int cmpres;
  struct stat s;
  
  buf[0] = '\0';

  /* 
   * if we are on a TU ASE server then
   *    we want to mess with something like /cc/st1/.ses.exports.inc
   *    which is actually an include file.  
   * else
   *    we want to mess with /etc/exports directly
   *    linux doesn't do include files in exports.
   */

  if ( stat("/var/ase/mnt",&s) || ( ! (s.st_mode & S_IFDIR) ) ) {
    strcpy(lockfile, "/etc");
    strcat(lockfile, "/exports.lock");

    strcpy(expfile, "/etc");
    strcat(expfile, "/exports");

    strcpy(newfile, "/etc");
    strcat(newfile, "/exports.new");
  }
  else {
    strcpy(lockfile, servmntpt);
    strcat(lockfile, "/.ses.exports.lock");

    strcpy(expfile, servmntpt);
    strcat(expfile, "/.ses.exports.inc");

    strcpy(newfile, servmntpt);
    strcat(newfile, "/.ses.exports.new");
  }

  if ( getlock(lockfile)) {
    syslog(LOG_ERR,"get lock failed %m");
    return(-1);
  }
  
  /* 
   * open the exports file, potentially creating it, 
   * and the exports.new file, definitely creating it
   * writing the new file as we go, 
   * read until we match on hostdir 
   * bunching split lines up into one rec
   * if not already there, insert the new ip in this rec
   *			   and split it up and write it
   *                       and copy across the rest of the file
   *                       and rename the new file, over the old file.
   * else just unlink the new file 
   * and return
   */

  f = fopen(expfile,"r");
  if ( ( f == (FILE *) NULL ) && ( errno == ENOENT ) ) {
    f = fopen(expfile, "a");
    fclose(f);
    f = fopen(expfile, "r" );
  }
  if ( f == (FILE *) NULL ) {
    droplock(lockfile);
    return(-1);
  }
  nf = fopen(newfile,"a");
  if (nf == (FILE *) NULL ) {
    fclose(f);
    droplock(lockfile);
    return(-1);
  }

  /* do this in alphabetical order */

  /* find our record */
  cmpres = -2;
  strcat(hostdir," ");
  while ( (! feof(f) ) && (fgets( buf, 1024, f) ) ) {
    if ( strlen(buf) > 0 )  {
      if ( (buf[0]==' ')  || 
           (buf[0]=='\n') || 
	   (buf[0]=='#')  ||
           ( (cmpres = strncmp(buf, hostdir, strlen(hostdir))) < 0) ){
        if (fputs(buf,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
      }
      else {
	break;
      }
    }
  }
  hostdir[strlen(hostdir)-1]='\0';

  /* if we did find this dir, then read entire record */
  if ( cmpres == 0){
    buf[strlen(buf)-1]='\0';
    strcpy(old,buf);
    while ( (! feof(f) ) && (fgets( buf, 1024, f) ) ) {
      if ( ( strlen(buf) > 0 ) && ( buf[0] == ' ' )) {
        buf[strlen(buf)-1]='\0';
	strcat(old,buf);
      }
      else {
        break;
      }
    }
    /* 
     * if our IP is in old, we can unlock and return 
     * else we add our ip to old, write it and carry on
     */
    p = old;
    q = p;
    /* find first ip */
    while ( (*q != '\0' ) && (*q != ' ') ) q++;
    while ( *q == ' ' ) q++;
    while ( *q == '-' ) {
      while ( (*q != '\0' ) && (*q != ' ') ) q++;
      while ( *q == ' ' ) q++;
    }
    strncpy(new,p,q-p);
    new[q-p] = '\0';
    p = q;
    while( strlen(p) > 0 ){
      while ( (*q != '\0' ) && (*q != ' ') ) q++;
      strncpy(ipsav,p,q-p);
      ipsav[q-p] = '\0';
      if ( *q == ' ' ) q++;
      if ( strcmp(ipsav,ip) < 0 ) {
	strncat(new,p,q-p);
	p = q;
      }
      else if (strcmp(ipsav,ip) == 0 ) {
	/* our ip is already in the list.  we are done */
/*  	printf("already there\n"); */
  	fclose(f);
        fclose(nf);
	unlink(newfile);
	droplock(lockfile);
	return(0);
      }
      else {
	break;
      }
    }
    /* write our ip, then this one, then carry on */
    if ( new[strlen(new)-1] != ' ') strcat(new," ");
    strcat(new,ip);
    if ( strlen(p) ){
      strcat(new," ");
      strncat(new,p,q-p);
      p = q;
      /* now copy over rest of IPs */
      while( strlen(p) > 0 ){
        while ( (*q != '\0' ) && (*q != ' ') ) q++;
        if ( *q == ' ' ) q++;
        strncat(new,p,q-p);
        p = q;
      }
    }

    strcat(new,"\n");
    strcpy(old,new);
    /* write it out */
    if ( strlen(old) < 80 ){
      /* just write it */
      if (fputs(old,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
    }
    else {
      /* write a bit, then a bit more etc */
      p = old;
      while ( strlen(p) > 80 ) {
        /* try and break the line near 60 chars */
        q = p + 60;
        while ( (*q != '\0' ) && (*q != ' ') ) q++;
        if ( (q-p) > 80 ) {
	  q--;
	  while ( ( q > p ) && (*q != ' ') ) q--;
	}
	if (q == p) {
	  q++;
	  while ( (*q != '\0' ) && (*q != ' ') ) q++;
	}
	if (*q == '\0'){
	  *q++ ='\n';
	  *q ='\0';
	  if (fputs(p,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
	  p = q;
	}
	else {
	  *q++ = '\n';
	  csave = *q;
	  *q = '\0';
  	  if (fputs(p,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
	  *q-- = csave;
	  *q = ' ';
	  p = q;
	}
      }
      if ( strlen(p) > 0 ){
        if (fputs(p,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
      }
    }
    /* in reading rest of record we may have filled buf */
    if ( strlen(buf) > 0 ) if (fputs(buf,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
  }
  else {
    /* 
     * we didn't find our directory, so create an entry for it, 
     * write it and carry on
     */
    strcpy(new,hostdir);
    strcat(new," ");
    /* if we are an ase cluster need to do this -m stuff */
    if ( (!stat("/var/ase/mnt",&s))&&(s.st_mode & S_IFDIR) ) {
      strcat(new,"-m=/var/ase/mnt/");
      strcat(new,host);
      strcat(new,hostdir);
    }
    strcat(new," ");
    strcat(new,ip);
    strcat(new,"\n\n"); /* blank line between entries */
    if (fputs(new,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
  }

  /* buf may contain the next entry after ours */
  if ( cmpres > 0) {
    if (fputs(buf,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
  }

  /* copy rest of file over */
  while ( (! feof(f) ) && (fgets( buf, 1024, f) ) ) {
    if ( strlen(buf) > 0 ) {
      if (fputs(buf,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
    }
  }
  fclose(f);
  fclose(nf);
  unlink(expfile);
  if ( link(newfile,expfile) ) {
    droplock(lockfile);
    return(-1);
  }
  unlink(newfile);
  if ( droplock(lockfile) ) {
    syslog(LOG_ERR,"drop lock failed %m");
    return(-1);
  }

  /* 
   * we have added our entry to the exports file.  
   * so we need to hup a daemon, but since there are potentially 
   * several mounts off the one daemon, we just pass a list of daemons to hup
   * back up to worker which hups all in the list once each after everything
   * else is finished.
   */
  if ( ! strlen(huplist) ) {
    strcat(huplist, host);
  }
  else {
    strcat(huplist," ");
    strcat(host, " ");
    if ( ! strstr(huplist,host) ){
      strcat(huplist, host);
    }
    huplist[strlen(huplist)] = '\0';
    host   [strlen(host   )] = '\0';
  }

  return(0);

}

int fixexpire( char *name, char *ip, char *dir, char *servmntpt) {
 
  char lockfile[128], expfile[128], newfile[128];
  FILE *f, *nf;
  char buf[1024], old[1024], new[1024];
  time_t now;
  struct stat s;
  
  buf[0] = '\0';

  /* 
   * if we are on a TU ASE server then
   *    we want to mess with something like /cc/st1/.ses.exports.inc
   *    which is actually an include file.  
   * else
   *    we want to mess with /etc/exports directly
   *    linux doesn't do include files in exports.
   */

  if ( stat("/var/ase/mnt",&s) || ( ! (s.st_mode & S_IFDIR) ) ) {
    strcpy(lockfile, "/etc");
    strcat(lockfile, "/expire.lock");

    strcpy(expfile, "/etc");
    strcat(expfile, "/expire");

    strcpy(newfile, "/etc");
    strcat(newfile, "/expire.new");
  }
  else {
    strcpy(lockfile, servmntpt);
    strcat(lockfile, "/.ses.expire.lock");

    strcpy(expfile, servmntpt);
    strcat(expfile, "/.ses.expire");

    strcpy(newfile, servmntpt);
    strcat(newfile, "/.ses.expire.new");

  }

  now = time((time_t *) 0);
  sprintf(new,"%s %s %d\n", dir, ip, now);

  if ( getlock(lockfile)) {
    syslog(LOG_ERR,"get lock failed %m");
    return(-1);
  }
  
  /* 
   * open the expiry file, potentially creating it, 
   * and the expiry.new file, definitely creating it
   * writing the new file as we go, 
   * read until we match on dir and ip
   * insert or replace the new record,
   * copy across the rest of the file
   * rename the new file, over the old file.
   */

  f = fopen(expfile,"r");
  if ( ( f == (FILE *) NULL ) && ( errno == ENOENT ) ) {
    f = fopen(expfile, "a");
    fclose(f);
    f = fopen(expfile, "r+" );
  }
  if ( f == (FILE *) NULL ) {
    droplock(lockfile);
    return(-1);
  }
  nf = fopen(newfile,"a");
  if (nf == (FILE *) NULL ) {
    fclose(f);
    droplock(lockfile);
    return(-1);
  }
  /* I'm doing this in alphabetical order */
  while ( (! feof(f) ) && (fgets( buf, 1024, f) ) ) {
    if ( ( strlen(buf) > 1 ) && ( buf[0] != '#' )) {
      if ( strncmp(buf, new, strlen(dir)+strlen(ip)+2) < 0){
	if (fputs(buf,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
      }
      else { 
	break;
      }
    }
  }
  if (fputs(new,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
  if ( ( strlen(buf) > 1 ) && ( buf[0] != '#' )) {
    if (strncmp(buf, new, strlen(dir)+strlen(ip)+2) > 0 ){
      if (fputs(buf,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
    }
  }
  /* copy rest of file over */
  while ( (! feof(f) ) && (fgets( buf, 1024, f) ) ) {
    if ( ( strlen(buf) > 1 ) && ( buf[0] != '#' )) {
      if (fputs(buf,nf)<0){fclose(f);fclose(nf);droplock(lockfile);return(-1);}
    }
  }
  fclose(f);
  fclose(nf);
  unlink(expfile);
  if ( link(newfile,expfile) ) {
    droplock(lockfile);
    return(-1);
  }
  unlink(newfile);
  if ( droplock(lockfile) ) {
    syslog(LOG_ERR,"drop lock failed %m");
    return(-1);
  }
  return(0);
}

int getmountinfo(char *dir, char *host, char *hostdir, char *servmntpt) {

#ifdef linux
    struct stat s;
#endif
#ifdef __osf__
    struct statfs s;
#endif

    int dev;
    char buf[1024], *p, *sd, *smp;   /* serverdisk and servermountpoint */
    FILE *f;
    char myname[128];
   
#ifdef linux

    if ( stat(dir,&s) ){
        syslog(LOG_ERR,"getmountinfo %m");
        return(-1);
    }
    dev = s.st_dev;

    f = fopen("/proc/mounts", "r" );
    if ( f == (FILE *) 0 ) {
        syslog(LOG_ERR,"nfsses df %m");
        return(-1);
    }
    while ( (! feof(f) ) && (fgets( buf, 1024, f) ) ) {
      /* for non blank, non comment lines */
      if ( ( strlen(buf) > 1 ) && ( buf[0] != '#' )) {
          buf[strlen(buf)-1] = '\0';
          smp = index(buf, ' ');
          if ( smp == (char *)NULL ) continue;
          *smp++ = '\0';
          while( *smp == ' ') smp++;
          p = smp;
          while ( (*p != ' ') && (*p != '\0') ) p++;
          *p = '\0';      
          if ( stat(smp,&s) ){
              syslog(LOG_ERR,"getmountinfo2 %m");
              return(-1);
          }
          if ( dev == s.st_dev ) break;
      }
    }

    if ( dev != s.st_dev ) return(-1);
    strcpy(servmntpt,smp);

#else

    if ( statfs(dir,&s) ){
        syslog(LOG_ERR,"statfs %m");
        exit();
    }

    strcpy(servmntpt,s.f_mntonname);
    strcpy(buf,s.f_mntfromname);

#endif

    /* if its local its easy */
    if ( ! strncmp(buf,"/dev/",5) ) {
	gethostname(myname,sizeof(myname));
        strcpy(host,myname);
	strcpy(hostdir,dir);
	return(0);
    }

    /* non local, assume nfs */
    /* given server:servdisk servmntpt or servdisk@server*/
    /* we want server:servdisk+(dir-servmntpt) dir nfs rw 0 0 */

    sd = index(buf,':');

    if ( sd != (char *)NULL) {
      *sd = '\0';
      sd++;
      strcpy(host,buf);

      strcpy(hostdir,sd);
      p = dir + strlen(servmntpt);
      strcat(hostdir,p);
      return(0);
    }

    sd = index(buf,'@');

    if ( sd != (char *)NULL) {
      *sd = '\0';
      sd++;
      strcpy(hostdir,buf);
      p = dir + strlen(servmntpt);
      strcat(hostdir,p);

      strcpy(host,sd);
      return(0);
    }
    

    return(-1);
    
}
int droplock(char *lf){
  char  apid[10], buf[10];
  pid_t  pid;
  int i,f;


  pid = getpid();
  sprintf(apid,"%d",pid);
  f = open(lf,O_RDONLY);
  if ( f < 0 ) {
    syslog(LOG_ERR,"unlocker1 %m");
    return(-1);
  }
  buf[0] = '\0';
  i = read(f,buf,10);
  if ( i < 0 ) {
    syslog(LOG_ERR,"unlocker2 %m");
    return(-1);
  }
  close(f);
  if ( strncmp(apid,buf,strlen(apid))){
    syslog(LOG_ERR," big stuffup.  pid in lockfile is %s, should be %s\n",buf,apid);
    return(-1);
  }
  
  if ( unlink(lf) ){
    syslog(LOG_ERR," big stuffup.  can't unlink lockfile\n");
  }
  return(0);
}


int getlock( char *lf ){
  char   pf[128], apid[10], buf[10];
  pid_t  pid;
  int    i, f, count;


  pid = getpid();
  sprintf(apid,"%d",pid);
  strcpy(pf,lf);
  strcat(pf,".");
  strcat(pf,apid);
AGAIN:
  f = open(pf,O_CREAT|O_EXCL|O_WRONLY,0700);
  if ( f < 0 ) {
    syslog(LOG_ERR,"locker1 %m");
    return(-1);
  }
  strcat(apid,"\n");
  i = write(f, apid, strlen(apid));
  if ( i < 0 ) {
    syslog(LOG_ERR,"locker2 %m");
    return(-1);
  }
  close(f);
  count = 10;
  while ( ( (i = link(pf,lf)) == -1 ) && (errno == EEXIST) && count-- ) sleep(1);
  if ( i < 0 ) {
    syslog(LOG_ERR,"locker3 %m");
    unlink(pf);
    return(-1);
  }
  if ( unlink(pf) ){
    syslog(LOG_ERR," little stuffup.  can't unlink pidfile\n");
  }
  f = open(lf,O_RDONLY);
  if ( f < 0 ) {
    syslog(LOG_ERR,"locker4 %m");
    return(-1);
  }
  i = read(f,buf,10);
  if ( i < 0 ) {
    syslog(LOG_ERR,"locker5 %m");
    return(-1);
  }
  close(f);
  if ( strncmp(apid,buf,strlen(apid))){
    goto AGAIN;
  }
  return(0);
}
